
module FFJ.Adapter where {
	import FFJ.Syntax;
	import FFJ.Program;
	import FFJ.Parser;
	
	getFeature :: (CT,PT) -> Feature -> [PStat] -> (CT,PT);
	getFeature (ct,pt) feat (x:xs) = getFeature (readStat (ct,pt) feat x) feat xs;
	getFeature (ct,pt) feat [] = (ct,pt);
	
	readStat :: (CT,PT) -> Feature -> PStat -> (CT,PT);
	readStat (ct,pt) feat (PClass c d pfields pmeth) = (ct++[((c,feat),(False,c,(getType d),(getFields pfields),(getMethods pmeth)))], pt);
	readStat (ct,pt) feat (PRef c pfields pmeth) = (ct++[((c,feat),(True,c,"Object",(getFields pfields),(getMethods pmeth)))], pt);
	readStat (ct,pt) feat (PDef c x t) = (ct,pt++[(((getType c),(x,feat)), (getTerm t))]);
	
	getType :: PCID -> Type;
	getType PObject = "Object";
	getType (PCID d) = d;

	getFields :: [PField] -> [Field];
	getFields ((PField (PDecl c f)):xs) = ((getType c),f) : (getFields xs);
	getFields [] = [];
	
	getMethods :: [PMethod] -> [Method];
	getMethods ((PMethod o (PDecl r m) ps t):xs) = (o,(getType r),m,(getParams ps),(getTerm t)) : (getMethods xs);
	getMethods [] = []; 
	
	getParams :: [PDecl] -> [Param];
	getParams ((PDecl t x):xs) = ((getType t), x) : (getParams xs);
	getParams [] = [];
	
	getTerm :: PTerm -> Term;
	getTerm (PVar PThis) = (Var "this"); 
	getTerm (PVar (PVID s)) =  (Var s);
	getTerm (PElim t ((PGet f):xs)) = embedTerm (Proj (getTerm t) f) xs;
	getTerm (PElim t ((PCall m ts):xs)) = embedTerm (Inv (getTerm t) m (map getTerm ts)) xs;
	getTerm (PElim t []) = getTerm t;
	getTerm (PCast pcid t) = (Cast (getType pcid) (getTerm t));
	getTerm (PNew pcid ts) = (Inst (getType pcid) (map getTerm ts));

	embedTerm :: Term -> [PElim] -> Term;
	embedTerm t ((PGet f):xs) = embedTerm (Proj t f) xs;
	embedTerm t ((PCall m ts):xs) = embedTerm (Inv t m (map getTerm ts)) xs;
	embedTerm t [] = t;
}	
