module FFJPL.TermTyping 
  ( getTermType )
where {
	import FFJ.Syntax;
	import FFJPL.Definitions;
	import FFJPL.FeatureModel;
	import FFJPL.Predicates;
	import FFJPL.Navigation;
	import FFJPL.Subtype;
	import FFJPL.FieldLookup;
	import FFJPL.MethodLookup;
	import FFJ.Print;
	import Data.List ( nub );
	import Debug.Trace;

    --
    -- Eliminates duplicates from the result list.
    --
    getTermType :: FeatureModel -> CT -> Gamma -> Term -> Feature -> [Type];
    getTermType fm ct gamma term phi =
    --  trace (show (length types',length types)) $
        types
        where {
          types' = getTermType' fm ct gamma term phi;
          types  = nub types';
        };
	
	getTermType' :: FeatureModel -> CT -> Gamma -> Term -> Feature -> [Type];
	getTermType' fm ct gamma (Var v) phi = case (lookup v gamma) of {
		Just c -> [c];
		Nothing -> error ("Type error: Type of local variable " ++ v ++ " is not known!");
	};

	getTermType' fm ct gamma (Proj t f) phi | (all (\e -> validRefField fm ct [phi] (e,f)) es) = [c | (c,g) <- (concat (concat (map (\xs -> fields fm ct [phi] xs) (map (getLast ct) es)))), g == f]
	                                       | otherwise = error ("Type error: Field access: " ++ (termToString (Proj t f)) ++ " is not well-formed!")
	                                       where { es = getTermType' fm ct gamma t phi; };

	getTermType' fm ct gamma (Inv t m ts) phi | (all (\e -> validRefMethod fm ct [phi] (e,m)) es) &&
	                                           (subTypesAllVariants fm ct css (fst (unzip sgns)) [phi]) = [b | (ds,b) <- sgns]
		                                     | otherwise = error ("Type error: Method invocation: " ++ (termToString (Inv t m ts)) ++ " is not well-formed!")
                                             where { es = getTermType' fm ct gamma t phi; sgns = ((concat (map (\xs -> mtype fm ct [phi] m xs) (map (getLast ct) es)))); css = (map (\t -> getTermType' fm ct gamma t phi) (ts));};

	getTermType' fm ct gamma (Inst c ts) phi | (validRef fm ct [phi] c) &&
	                                          (all (\ds -> length css == length ds) dss) &&
	                                          (subTypesAllVariants fm ct css dss [phi]) = [c]
	                                        | otherwise = error ("Type error: Object creation: " ++ (termToString (Inst c ts)) ++ " is not well-formed!")
	                                        where { css = (map (\t -> getTermType' fm ct gamma t phi) (ts)); dss = (map (map (fst)) (fields fm ct [phi] (getLast ct c))); };
	                                        
	getTermType' fm ct gamma (Cast c t) phi | validC && (all (\e -> (isSubType fm ct e c [phi]) || (isSubType fm ct c e [phi])) es) = [c]
	                                       | validC && (any (\e -> (not (isSubType fm ct c e [phi])) && (not (isSubType fm ct e c [phi]))) es) = trace ("Stupid warning in cast: " ++ (termToString(Cast c t)) ++ "!") [c]
	                                       | otherwise =  error ("Type error: Cast: " ++ (termToString (Cast c t)) ++ " is not well-formed!") 
	                                       where { es = getTermType' fm ct gamma t phi; validC = validRef fm ct [phi] c; };
}