
module FJ.Typing where {
	import FJ.Syntax;
	import FJ.Program;
	import FJ.AuxFunc;
	import List;
	import Maybe;
	import Debug.Trace;
	
	type Gamma = [(Variable,Type)];
	
	ctOK :: CT -> Bool;
	ctOK ct = (objectIsOK ct) && (objectIsUnique ct) && (hasNoCycles ct ct);

	objectIsOK :: CT -> Bool;
	objectIsOK ct = case (lookup "Object" ct) of {
		Just (c,d,k,f,m) -> ((length f == 0) && (length m == 0) && (d == ""));
		Nothing -> False; 
	};

	objectIsUnique :: CT -> Bool;
	objectIsUnique ct = (length (findIndices (== "Object") (fst (unzip ct))) == 1);

	hasNoCycles :: CT -> CT -> Bool;
	hasNoCycles ct ((c,cd):xs) = (not (inducesCycle ct c (superclass cd))) && hasNoCycles ct xs;
	hasNoCycles ct [] = True;
	
	inducesCycle :: CT -> Type -> Type -> Bool;
	inducesCycle ct c d | ((c == "Object") || (d == "Object")) = False
                        | (c == d) = True 
	                    | otherwise = if (isJust cd) then inducesCycle ct c sc else error ("Class declaration of " ++ c ++ " does not exist!")
	                    where { cd = (classdef ct d); sc = superclass (fromJust cd) };

	allOK :: CT -> PT -> Bool;
	allOK ct pt = (ctOK ct) && (classesOK ct ct) && (programsOK ct pt);
	
	programsOK :: CT -> PT -> Bool;
	programsOK ct (p:ps) = (programOK ct p) && (programsOK ct ps);
	programsOK ct [] = True;

	programOK :: CT -> (PTEntry, Term) -> Bool;
	programOK ct (_,t) = isJust (classdef ct (termType ct [] t));

	classesOK :: CT -> CT -> Bool;
	classesOK ct ((_,c):cs) = (classOK ct c) && (classesOK ct cs);
	classesOK ct [] = True;
	
	classOK :: CT -> Class -> Bool;
	classOK ct (c,d,k,fs,ms) = (if (c /= "Object") then (consOK ct k c d fs) else True) && (methodsOK ct ms c d);
	
	consOK :: CT -> Cons -> Type -> Type -> [Field] -> Bool;
	consOK ct (k,ps,sargs,inits) c d fs = let ds = (fields ct d) in (k == c) && (equalLists (snd (unzip fs)) (fst (unzip inits))) && (equalLists (snd (unzip fs)) (snd (unzip inits))) && (isPrefixOf (snd ( unzip ds)) (snd (unzip ps))) && (isSuffixOf (snd (unzip fs)) (snd (unzip ps))) && (((length fs) + (length ds)) == (length ps));
	
	equalLists :: [String] -> [String] -> Bool;
	equalLists (a:as) (b:bs) = (a == b) && (equalLists as bs);
	equalLists [] [] = True;
	equalLists [] _ = False;
	equalLists _ [] = False;
	
	methodsOK :: CT -> [Method] -> Type -> Type -> Bool;
	methodsOK ct (m:ms) c d = (methodOK ct m c d) && (methodsOK ct ms c d);
	methodsOK _ [] _ _ = True;
	
	methodOK :: CT -> Method -> Type -> Type -> Bool;
	methodOK ct (r,m,ps,t) c d = let dd = (classdef ct d) in if (isJust dd) then (overrideClass ct m (fromJust dd) (fst (unzip ps),r)) && (subtype ct (termType ct ((zip (snd (unzip ps)) (fst (unzip ps)))++[("this",c)]) t) r) else error ("Type error: class declaration of " ++ d ++ " does not exist!");
	
	termType :: CT -> Gamma -> Term -> Type;
	termType ct gamma (Var v) = case (lookup v gamma) of {
		Just c -> c;
		Nothing -> error ("Type error: type of variable " ++ v ++ " not known!");
	};
	
	termType ct gamma (Proj t f) = getFieldType f (fields ct (termType ct gamma t));
	
	termType ct gamma (Inv t m ts) = let c0 = (termType ct gamma t) in case (mtype ct m c0) of {
		Just (ds,c) -> if (subtypes ct (map (termType ct gamma) ts) ds) then c else error ("Type error: signature of method " ++ m ++ " does not match!");
		Nothing -> error ("Type error: method " ++ m ++ " does not exist in class " ++ c0 ++ "!");
	};
	
	termType ct gamma (Inst c ts) = if (subtypes ct (map (termType ct gamma) ts) (fst (unzip (fields ct c)))) then c else error ("Type error: arguments do not match the fields of class " ++ c ++"!");
	
	termType ct gamma (Cast c t) | (subtype ct d c) = c
	                             | (subtype ct c d) && (c /= d) = c
	                             | otherwise = error ("Type error: stupid cast of type " ++ d ++ " to type " ++ c ++ "!")
	                             where { d = (termType ct gamma t); };
	
	--wellformed :: Method -> Bool;
	

}
